#include "convolutionS.h"
#include "gpuOperations.h"

#define a3d(a,x,y,z,xs,ys,zs) a[x*ys*zs + y*zs + z]

extern __shared__ cufftComplex shared[];

__global__ void convolutionS(	cufftComplex* d_Data, 
								cufftComplex* d_Result, 
 								cufftComplex* d_S,
 								const float mu, 
 								const int xsize, 
 								const int ysize, 
 								const int zsize, 
								const size_t pitch, 
								const int bandwidth, 
								const int fc) {
	
	const int bandwidthStart = fc - bandwidth / 2;
	
	const int x = threadIdx.x; //0-10
	const int y = threadIdx.y; //0-10
	
	const int xs = blockDim.x; //11
	const int ys = blockDim.y; //11
	
	const int blockStartx = blockIdx.x; //0-10
	const int blockStarty = blockIdx.y; //0-10
	
	const int globalx = x - KERNEL_RADIUS + blockStartx*xs - blockStartx*2*KERNEL_RADIUS;
	const int globaly = y - KERNEL_RADIUS + blockStarty*ys - blockStarty*2*KERNEL_RADIUS;
	
	const int smemPos = x + y*xs;
	const int B = xs*ys; //11*11 = 121

	//Set shared memory:
//	__shared__ cufftComplex data[11*11*(BANDWIDTH+1)]; //(11*11*15) //Including aprons
//    __shared__ cufftComplex s[11*11]; //11*11 //(10*10) //Including aprons
    
    cufftComplex *data = shared;
    cufftComplex *s = &(shared[xs*ys*(bandwidth+1)]);
	
	//Optimization, uses less registers
	bool active = false;
	bool write = false;
	if(globalx >=  0 && globaly >= 0 && globalx <= xsize-1 && globaly <= ysize-1) {
		active = true;
		if ((x >= KERNEL_RADIUS && x <= ((xs-1)-KERNEL_RADIUS) && y >= KERNEL_RADIUS && y <= ((ys-1)-KERNEL_RADIUS))) {
			write = true;
		}
	}
	
	//If thread inside matrix
	if (active) {
		//Load S matrix:
		//load data
		cufftComplex *element = (cufftComplex*) ((char*)d_S + globaly*pitch) + globalx;
		s[smemPos].x = element->x;
		s[smemPos].y = element->y;
		
		//Load 3D data:
		//for loop all planes
		#pragma unroll 15
		for (int i = 0; i < (bandwidth+1); i++) {
			//load data
			data[smemPos + i*B].x = a3d(d_Data, globalx, globaly, bandwidthStart+i, xsize, ysize, zsize).x;
			data[smemPos + i*B].y = a3d(d_Data, globalx, globaly, bandwidthStart+i, xsize, ysize, zsize).y;
		}
	}
	//else
	else {
		//set to zero
		s[smemPos].x = 0.0f;
		s[smemPos].y = 0.0f;
		//for loop all planes
		#pragma unroll 15
		for (int i = 0; i < (bandwidth+1); i++) {
			//set to zero
			data[smemPos + i*B].x = 0.0f;
			data[smemPos + i*B].y = 0.0f;
		}
	}
	
	//sync
	__syncthreads();
	
	//Compute:
	cufftComplex sum;
	sum.x = 0.0f;
	sum.y = 0.0f;
	//if thread inside matrix == if pos-kernel_w > outside lower block and pos+kernel_w < outside upper block
	if (write) {
		cufftComplex R_klmn, W_klmn, M_klmn;
		float R_kl, R_mn;
		
		//R(k,l)
		R_kl = 0.0f;
		#pragma unroll 15
		for (int i = 0; i < bandwidth+1; i++) {
			R_kl += 	data[smemPos+i*B].x	*	data[smemPos+i*B].x
					+ 	data[smemPos+i*B].y	*	data[smemPos+i*B].y;
		}
		R_kl /= (bandwidth+1.0f);
		R_kl = sqrtf(R_kl);
		
		//for loop vertical k
		#pragma unroll 5
		for (int k = -KERNEL_RADIUS; k <= KERNEL_RADIUS; k++) {
			//for loop horizontal l
			#pragma unroll 5
			for (int l = -KERNEL_RADIUS; l <= KERNEL_RADIUS; l++) {
				//compute new S
				R_mn = 0.0f;
				R_klmn.x = 0.0f;
				R_klmn.y = 0.0f;
				W_klmn.x = 0.0f;
				W_klmn.y = 0.0f;
				M_klmn.x = 0.0f;
				M_klmn.y = 0.0f;
				
				//R(k,l,m,n)
				#pragma unroll 15
				for (int i = 0; i < bandwidth+1; i++) {
					R_klmn.x +=  		data[smemPos+i*B].x	*	data[smemPos + k*xs + l + i*B].x
									+ 	data[smemPos+i*B].y	*	data[smemPos + k*xs + l + i*B].y;
					R_klmn.y +=  		data[smemPos+i*B].y	*	data[smemPos + k*xs + l + i*B].x
									- 	data[smemPos+i*B].x	*	data[smemPos + k*xs + l + i*B].y;
				}
				R_klmn.x /= (bandwidth+1.0f);
				R_klmn.y /= (bandwidth+1.0f);
				
				//R(k-m,l-n)
				#pragma unroll 15
				for (int i = 0; i < bandwidth+1; i++) {
					R_mn +=  	data[smemPos + k*xs + l + i*B].x	*	data[smemPos + k*xs + l + i*B].x
							+ 	data[smemPos + k*xs + l + i*B].y	*	data[smemPos + k*xs + l + i*B].y;
				}
				R_mn /= (bandwidth+1.0f);
				R_mn = sqrtf(R_mn);
				
				//W(k,l,m,n)
				W_klmn.x = R_klmn.x / (R_kl * R_mn + 1e-5);
				W_klmn.y = R_klmn.y / (R_kl * R_mn + 1e-5);
				
				//M(k,l,m,n)
				M_klmn.x = sqrtf(W_klmn.x*W_klmn.x + W_klmn.y*W_klmn.y) * R_klmn.x;
				M_klmn.y = sqrtf(W_klmn.x*W_klmn.x + W_klmn.y*W_klmn.y) * R_klmn.y; 								
				
				//new_s
				sum.x += (M_klmn.x * s[smemPos + k*xs + l].x + M_klmn.y * s[smemPos + k*xs + l].y) / (s[smemPos + k*xs + l].x*s[smemPos + k*xs + l].x + s[smemPos + k*xs + l].y*s[smemPos + k*xs + l].y + 1e-5);
				sum.y += (M_klmn.y * s[smemPos + k*xs + l].x - M_klmn.x * s[smemPos + k*xs + l].y) / (s[smemPos + k*xs + l].x*s[smemPos + k*xs + l].x + s[smemPos + k*xs + l].y*s[smemPos + k*xs + l].y + 1e-5);
			}
		}
		//calculate new S element
		sum.x = (1.0f - mu) * s[smemPos].x + mu * sum.x;
		sum.y = (1.0f - mu) * s[smemPos].y + mu * sum.y;
		
		//write s to memory
		cufftComplex *element = (cufftComplex*) ((char*)d_Result + globaly*pitch) + globalx;
		element->x = sum.x;
		element->y = sum.y;
	}
}
