#include "cpuCorrelation.h"
#include "gpuOperations.h"

/*
 * Computes the correlation by running a filter over elementMatrix, 
 * this filter multiplies the center point of the filter with the 
 * complex conjugate of all other points being filtered and sums 
 * them up. The result is stored in weightMatrix.
 */
void correlate(	cufftComplex **elementMatrix, 
				cufftComplex ***correlationMatrix, 
				cufftComplex ***weightMatrix, 
				int xsize, 
				int ysize, 
				int paddedxsize, 
				int paddedysize) {
	
	int i, j, k, l, xoffset, yoffset;
	cufftComplex sum, yMULyconv;
	float ABSyMULycnsq, ABSysq, ABSycnsq;
	xoffset = (paddedxsize-xsize)/2;
	yoffset = (paddedysize-ysize)/2;
	for (i = xoffset; i < xsize+xoffset; i++) {
		for (j = yoffset; j < ysize+yoffset; j++) {
			for (k = i-xoffset; k < i+xoffset; k++) {
				for (l = j-yoffset; l < j+yoffset; l++) {
					//Center multiplied with complex conjugate of values inside grid.
//					printf("i & j (%d,%d) k & l: (%d,%d)\n",i,j,k,l);
					/*ORIGINAL CODE
					correlationMatrix[0][k-(i-xoffset)][l-(j-yoffset)].x = 
							elementMatrix[i][j].x * complexConjugate(elementMatrix[k][l]).x;
							- elementMatrix[i][j].y * complexConjugate(elementMatrix[k][l]).y;
					correlationMatrix[0][k-(i-xoffset)][l-(j-yoffset)].y = 
							elementMatrix[i][j].x * complexConjugate(elementMatrix[k][l]).y
							+ elementMatrix[i][j].y * complexConjugate(elementMatrix[k][l]).x;
					*/		
					yMULyconv.x =  	elementMatrix[i][j].x	*	elementMatrix[k][l].x
            						- elementMatrix[i][j].y	*	complexConjugate(elementMatrix[k][l]).y;
            		yMULyconv.y =	elementMatrix[i][j].x	*	complexConjugate(elementMatrix[k][l]).y
            						+ elementMatrix[i][j].y	*	elementMatrix[k][l].x;
            
            		//TODO check if this square root should really be here
					ABSyMULycnsq = 	/*sqrtf(*/	sqrtf(		yMULyconv.x 	* yMULyconv.x 
															+ yMULyconv.y 	* yMULyconv.y) 
												* 	sqrtf(		yMULyconv.x 	* yMULyconv.x 
															+ yMULyconv.y 	* yMULyconv.y);//);
							
					ABSysq =		sqrtf(		sqrtf(			elementMatrix[i][j].x	 		* elementMatrix[i][j].x 
																+ 	elementMatrix[i][j].y 		* elementMatrix[i][j].y) 
												* sqrtf(		elementMatrix[i][j].x 			* elementMatrix[i][j].x 
																+ 	elementMatrix[i][j].y 		* elementMatrix[i][j].y));
										
					ABSycnsq = 		sqrtf(		sqrtf(			elementMatrix[k][l].x 		* elementMatrix[k][l].x 
																+	elementMatrix[k][l].y 	* elementMatrix[k][l].y) 
												* sqrtf(		elementMatrix[k][l].x 		* elementMatrix[k][l].x 
																+	elementMatrix[k][l].y 	* elementMatrix[k][l].y));
			
					if (!(ABSycnsq < 0.000000001 && ABSycnsq > -0.000000001) && !(ABSysq < 0.000000001 && ABSysq > -0.000000001))
					{
					correlationMatrix[0][k-(i-xoffset)][l-(j-yoffset)].x = 	(ABSyMULycnsq / (ABSysq * ABSycnsq))
						* (		(	yMULyconv.x 			* 1  	//a*c
								+ 	yMULyconv.y 			* 0) 	//+ b*d
							/ 	(	1 	* 1  	//c*c
								+ 	0 	* 0));	//+ d*d*/
					correlationMatrix[0][k-(i-xoffset)][l-(j-yoffset)].y = 	(ABSyMULycnsq / (ABSysq * ABSycnsq)) 
						* (		(	yMULyconv.y 			* 1 	//b*c
								- 	yMULyconv.x 			* 0) 	//- a*d
							/ 	(	1 	* 1 	//c*c
								+ 	0 	* 0));	//+ d*d
					}
				}
			}
			//Unaveraged version:
//			weightMatrix[0][i-xoffset][j-yoffset].x = 0;//sum(correlationMatrix[0], (paddedxsize-xsize), (paddedysize-ysize));
			//Averaged version:
			//weightMatrix[0][i-xoffset][j-yoffset] = average(sum(correlationMatrix[0], (paddedxsize-xsize), (paddedysize-ysize)), (paddedxsize-xsize), (paddedysize-ysize));
		}
	}
}



void correlateReduce(	fftwf_complex ***fftM, 
						cufftComplex **sM, 
						cufftComplex **corrM, 
						int xsize, 
						int ysize, 
						int kernel_radius, 
						int kernel_wf, 
						float mu, 
						int bandwidth, 
						int fc) {
	
	//Compute:
	cufftComplex sum, R_klmn, W_klmn, M_klmn, data, dataNeighbor, s;
	int bwS = fc - bandwidth / 2; //bandwidthStart
	float R_kl, R_mn, bandwidthf;
	bandwidthf = bandwidth * 1.0f;
	//for loop x-direction
	for (int x = 0; x < xsize; x++) {
		//for loop y-direction
		for (int y = 0; y < ysize; y++) {
			sum.x = 0.0f;
			sum.y = 0.0f;
			
			//R(k,l)
			R_kl = 0.0f;
			for (int i = 0; i < bandwidth+1; i++) {
				data.x = fftM[x][y][bwS+i][0];
				data.y = fftM[x][y][bwS+i][1];
				R_kl += 	data.x	*	data.x
						+ 	data.y	*	data.y;
			}
			R_kl /= (bandwidthf+1.0f);
			R_kl = sqrtf(R_kl);
			
			//for loop vertical k
			for (int k = -kernel_radius; k <= kernel_radius; k++) {
				//for loop horizontal l
				for (int l = -kernel_radius; l <= kernel_radius; l++) {
					//compute new S
					R_mn = 0.0f;
					R_klmn.x = 0.0f;
					R_klmn.y = 0.0f;
					W_klmn.x = 0.0f;
					W_klmn.y = 0.0f;
					M_klmn.x = 0.0f;
					M_klmn.y = 0.0f;
					//R(k,l,m,n)
					for (int i = 0; i < bandwidth+1; i++) {
						data.x = fftM[x][y][bwS+i][0];
						data.y = fftM[x][y][bwS+i][1];
						if (x+l >= 0 && x+l < xsize && y+k >= 0 && y+k < ysize) {
							dataNeighbor.x = fftM[x+l][y+k][bwS+i][0];
							dataNeighbor.y = fftM[x+l][y+k][bwS+i][1];
						}
						else {
							dataNeighbor.x = 0.0f;
							dataNeighbor.y = 0.0f;
						}
						R_klmn.x +=  		data.x	*	dataNeighbor.x
										+ 	data.y	*	dataNeighbor.y;
						R_klmn.y +=  		data.y	*	dataNeighbor.x
										- 	data.x	*	dataNeighbor.y;
					}
					R_klmn.x /= (bandwidthf+1.0f);
					R_klmn.y /= (bandwidthf+1.0f);
					
					//R(k-m,l-n)
					for (int i = 0; i < bandwidth+1; i++) {
						if (x+l >= 0 && x+l < xsize && y+k >= 0 && y+k < ysize) {
							dataNeighbor.x = fftM[x+l][y+k][bwS+i][0];
							dataNeighbor.y = fftM[x+l][y+k][bwS+i][1];
						}
						else {
							dataNeighbor.x = 0.0f;
							dataNeighbor.y = 0.0f;
						}
						R_mn +=  	dataNeighbor.x	*	dataNeighbor.x
								+ 	dataNeighbor.y	*	dataNeighbor.y;
					}
					R_mn /= (bandwidthf+1.0f);
					R_mn = sqrtf(R_mn);
					
					//W(k,l,m,n)
					W_klmn.x = R_klmn.x / (R_kl * R_mn + 1e-5);
					W_klmn.y = R_klmn.y / (R_kl * R_mn + 1e-5);
					
					//M(k,l,m,n)
					M_klmn.x = sqrtf(W_klmn.x*W_klmn.x + W_klmn.y*W_klmn.y) * R_klmn.x;
					M_klmn.y = sqrtf(W_klmn.x*W_klmn.x + W_klmn.y*W_klmn.y) * R_klmn.y; 								
					
					//new_s
					if (x+l >= 0 && x+l < xsize && y+k >= 0 && y+k < ysize) {
						s.x = sM[x+l][y+k].x;
						s.y = sM[x+l][y+k].y;
					}
					else {
						s.x = 0.0f;
						s.y = 0.0f;
					}
					sum.x += (M_klmn.x * s.x + M_klmn.y * s.y) / (s.x*s.x + s.y*s.y + 1e-5);
					sum.y += (M_klmn.y * s.x - M_klmn.x * s.y) / (s.x*s.x + s.y*s.y + 1e-5);
				}
			}
			
			sum.x = (1.0f - mu) * sM[x][y].x + mu * sum.x;
			sum.y = (1.0f - mu) * sM[x][y].y + mu * sum.y;
			
			//Write new correction matrix to memory:
			corrM[x][y].x = sum.x;
			corrM[x][y].y = sum.y;
		}
	}
}
